/******************************************************************************
  (c) Copyright 2002,2003, 1060 Research Ltd                                   

  This Software is licensed to You, the licensee, for use under the terms of   
  the 1060 Public License v1.0. Please read and agree to the 1060 Public       
  License v1.0 [www.1060research.com/license] before using or redistributing   
  this software.                                                               

  In summary the 1060 Public license has the following conditions.             
  A. You may use the Software free of charge provided you agree to the terms   
  laid out in the 1060 Public License v1.0                                     
  B. You are only permitted to use the Software with components or applications
  that provide you with OSI Certified Open Source Code [www.opensource.org], or
  for which licensing has been approved by 1060 Research Limited.              
  You may write your own software for execution by this Software provided any  
  distribution of your software with this Software complies with terms set out 
  in section 2 of the 1060 Public License v1.0                                 
  C. You may redistribute the Software provided you comply with the terms of   
  the 1060 Public License v1.0 and that no warranty is implied or given.       
  D. If you find you are unable to comply with this license you may seek to    
  obtain an alternative license from 1060 Research Limited by contacting       
  license@1060research.com or by visiting www.1060research.com                 

  NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC  
  LICENSE V1.0 FOR DETAILS                                                     

  THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ     
  THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license      

  File:          $RCSfile: XSLTAccessor.java,v $
  Version:       $Name:  $ $Revision: 1.11 $
  Last Modified: $Date: 2008/05/02 12:17:25 $
 *****************************************************************************/
package org.netkernel.xml.accessor.xslt;

import org.apache.xalan.transformer.TrAXFilter;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.impl.ActiveEndpointMetaBuilder;
import org.netkernel.layer0.meta.impl.MetaRepresentationImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.util.*;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.urii.IIdentifier;
import org.netkernel.xml.representation.IRepSAX;
import org.netkernel.xml.representation.RepSAXFilter;
import org.netkernel.xml.util.*;


import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.sax.*;

import org.w3c.dom.Document;
import org.xml.sax.*;
import org.xml.sax.helpers.*;


import java.io.*;
/**
 *
 * @author  pjr
 */
public class XSLTAccessor extends StandardAccessorImpl
{
	public XSLTAccessor()
	{	this.declareThreadSafe();
		this.declareName("XSLT Transform");
		this.declareSourceRepresentation(String.class);
	}

	public void onSource(INKFRequestContext aContext) throws Exception
	{	
		// get transform
		INKFRequestReadOnly thisReq=aContext.getThisRequest();
		boolean isCompiled = thisReq.getArgumentValue(ActiveEndpointMetaBuilder.ARG_ACTIVE_TYPE).equals("xsltc");
		Class transformerAspectClass = isCompiled ? RepCompiledXSLTransformer.class : RepXSLTransformer.class;
		RepXSLTransformer ta = (RepXSLTransformer)aContext.source("arg:operator",transformerAspectClass);
		
		// create xsl parameter list
		PairList params = new PairList(1);
		OrderedCheapMultiMap paramMap = ta.getParameters();
		for (int i=0; i<paramMap.size(); i++ )
		{	String paramName=(String)paramMap.getKey(i);
			String arg=aContext.getThisRequest().getArgumentValue(paramName);
			Class argType = (Class)paramMap.getValue(i);
			if(arg!=null)		//Tolerant of missing params - let grammar binding enforce constraints
			{	Object paramObject = aContext.source(arg,argType);
				params.put(paramName, paramObject);
			}
		}
			
		// choose the most efficient way to get the input into XSLT
		Source inputSource=null;
		INKFResponse response=null;
		
		INKFResponseReadOnly operandResp = aContext.sourceForResponse("arg:operand");
		Object operandRep=operandResp.getRepresentation();
		if (operandRep instanceof Document)
		{	Document d = (Document)operandRep;
			inputSource = new DOMSource(d);
		}
		else if(operandRep instanceof IHDSNode || operandRep instanceof IHDSNodeList)
		{	Document d=(Document)aContext.source("arg:operand", Document.class);   //Do a source so that the transrepted HDS has chance to be cached
			inputSource=new DOMSource(d);
		}
		else if (operandRep instanceof IRepSAX)
		{ 	IRepSAX operand = (IRepSAX)operandRep;
			SAXTransformerFactory std= isCompiled ?
			(SAXTransformerFactory)org.apache.xalan.processor.TransformerFactoryImpl.newInstance() :
			(SAXTransformerFactory)org.apache.xalan.xsltc.trax.SmartTransformerFactoryImpl.newInstance();
			TrAXFilter filter=(TrAXFilter)std.newXMLFilter(ta.getTemplate());
			RepSAXFilter rep=new RepSAXFilter(filter, operand);
			Transformer f = filter.getTransformer();
			f.clearParameters();
			f.setURIResolver(new NKFURIResolver(aContext));
			for (int i=0; i<params.size(); i++)
			{	f.setParameter((String)params.getValue1(i),params.getValue2(i));
			}
			response = aContext.createResponseFrom(rep);
			response.setNoCache();		//Do not cache this filter - in SAX they are one shot and must be rebuilt each run of the pipeline
		}
		else 
		{	
			IReadableBinaryStreamRepresentation aspect = null;
			if(operandRep instanceof IReadableBinaryStreamRepresentation )
			{	aspect=(IReadableBinaryStreamRepresentation)operandRep;
			}
			else
			{	aspect=(IReadableBinaryStreamRepresentation)aContext.source("arg:operand",IReadableBinaryStreamRepresentation.class);
			}
			XMLReader aReader=XMLReaderFactory.createXMLReader();
			aReader.setEntityResolver(new NKFEntityResolver(aContext));
			if (aspect!=null)
			{	inputSource = new SAXSource(aReader, new InputSource(aspect.getInputStream()));
			}
			else
			{	//handle a null operand in a less-inelegant way
				StringReader r = new StringReader("<null/>");
				inputSource = new SAXSource(aReader, new InputSource(r));
			}
		}
		
		//if we have not generated a streaming aspect then do the work here with the Source
		if (inputSource!=null)
		{	// choose result type
			Result result = null;
			Class repClass=aContext.getThisRequest().getRepresentationClass();
			boolean isString=false, isStream=false;
			Object rep=null;
			if(String.class.isAssignableFrom(repClass))
			{	isString=true;
				rep = new StringWriter(256);
				result=new StreamResult((StringWriter)rep);
			}
			else if(IBinaryStreamRepresentation.class.isAssignableFrom(repClass))
			{	isStream=true;
				rep=new ByteArrayOutputStream(256);
				result=new StreamResult(new OutputStreamWriter((ByteArrayOutputStream)rep,"UTF-8"));
			}
			else
			{	result=new DOMResult();	
			}
			//////////////////////
			//perform transformation
			//////////////////////
			Transformer t = ta.getTemplate().newTransformer();
			for (int i=0; i<params.size(); i++)
			{	t.setParameter((String)params.getValue1(i),params.getValue2(i));
			}
			NKFURIResolver resolver = new NKFURIResolver(aContext);
			t.setURIResolver(resolver);
			try
			{	t.transform(inputSource, result);
				resolver.cleanup();
				//use appropriate response representation
				if(isStream)
				{	ByteArrayOutputStream baos=(ByteArrayOutputStream)rep;
					response=aContext.createResponseFrom(new ByteArrayRepresentation(baos.toByteArray(),"UTF-8"));
				}
				else if(isString)
				{	StringWriter sw=(StringWriter)rep;
					response=aContext.createResponseFrom(sw.getBuffer().toString());
				}
				else
				{	response = aContext.createResponseFrom((Document)((DOMResult)result).getNode());
				}
			}
			catch (TransformerException e)
			{	throw new NKFException("Unhandled Exception in XSLT",null,e.getLocationAsString(),e);
			}
		}
		
		response.setMimeType(ta.getMimeType());
	}
	
	@Override
	public void onMeta(INKFRequestContext aContext) throws Exception
	{	super.onMeta(aContext);
		if(aContext.isResponseSet())
		{	Object primary=aContext.getThisRequest().getPrimary();
			if(primary!=null)
			{	IIdentifier id=(IIdentifier)primary;
				IEndpointMeta meta=getMeta();
				OrderedCheapMultiStringMap map=meta.getIdentifierGrammar().parse(id.toString());
				String operator=map.get("operator");
				boolean isCompiled = map.get("type").equals("xsltc");
				Class transformerAspectClass = isCompiled ? RepCompiledXSLTransformer.class : RepXSLTransformer.class;
				RepXSLTransformer ta = (RepXSLTransformer)aContext.source(operator,transformerAspectClass);
				OrderedCheapMultiMap params=ta.getParameters();
				StringBuilder sb=new StringBuilder(128);
				sb.append("<meta>");
				sb.append("<desc>Tolerant XSLT Parameters</desc>");
				sb.append("<params>");
				for(int i=0; i<params.size(); i++)
				{	sb.append("<param>");
					sb.append("<name>");
					String name=(String)params.getKey(i);
					sb.append(name);
					sb.append("</name>");
					sb.append("<class>");
					sb.append(((Class)params.getValue(i)).getName());
					sb.append("</class>");
					sb.append("</param>");
				}
				sb.append("</params>");
				sb.append("</meta>");
				MetaRepresentationImpl metaImpl=new MetaRepresentationImpl(id.toString(), sb.toString(), null, null);
				aContext.createResponseFrom(metaImpl);
			}
		}
	}
}